# Copyright (C) 2008, Polychronis Ypodimatopoulos
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import sys, time, os, re
from string import upper
from random	import random
from struct import pack, unpack
from math   import ceil, pi

from cerebro.node_struct import Node, MST
from cerebro.common import *

class MeshPresence():
    def __init__(self, nid, interval=5, updateCallBack=None):
        self.interval		= interval
        self.DEBUG			= 8
        self.T				= 30
        self.silent			= False
        self.HOP_WEIGHT		= 2
        
        # The following are index numbers for easy reference
        self.NID			= 0
        self.WITNESS		= 1
        self.DIST			= 2
        self.SEQN           = 3

        self.nid			= nid
        self.nidHex			= bin2mac(self.nid)
        self.nid_length		= len(nid)
        
        self.root           = Node(self.nid, par=None)
        
        # keeps track of my own sequence number
        self.sequenceno     = 0

        # keeps track of other nodes' sequence numbers
        self.seqnos         = {}

        # holds min distastances to all nodes in ptable
        # mindists['nodeid'] = [path_to_nodid]
        self.mst			= MST()
        
        self.noise_thres	= 0.0
        
        # updateCallBack is another way (besides HTTP)
        # for receiving updates from MeshPresence
        self.updateCallBack = updateCallBack
        
    def above_artificial_noise(self):
        res = 1
        if random() < self.noise_thres:
            res = 0
        return res
        
    def estimate_cutoff_exp(self, node):
        arrno = len(node.arr_times) # of arrivals
        # formulate a new arrival rate for this entry
        try:
            node.dist = 1.0*self.T / (arrno * self.interval)
        except:
            pout(0, "PANIC: Tried to estimate rate on a node with 0"+\
                "arrivals:", node.nid)
            raise

        # est = 2.3*T/arrivals
        node.arr_est = 2.3 * node.dist
        
#    def is_online(self, node):
#        """ if at least one entry is found for "node", where
#            the last arrival time is within the estimate, then node is online """
#        result = False
#        for entry in self.root.preorder():
#            if entry.nid == node:
#                lastarr  = entry.arr_times[-1] # last arrival
#                if now - entry.arr_est < lastarr:
#                    result = True
#        
#        return result
        
    def add_distances(self, d1, d2):
        if d1 and d2:
            return d1 * d2 * self.HOP_WEIGHT
        else:
            # if one of them is 0, just add them
            return d1 + d2
        
    def remove_subtree(self, node):
        for k in node.postorder():
            del k
        del node.parent.goto[node.nid]
        del node
        
    def parse_entries(self, load):
        # parse_entries breaks up the frame payload into an array of arrays
        # Each entry is 63 bytes: nid:6, witness:6, dist:2
        # Therefore, the size of load modulo 14 should be zero
        result = []
        ENTRY_SIZE = 2 * self.nid_length + 2 + 2
        if len(load) % ENTRY_SIZE:
            print "Panic: wrong payload size! %d bytes: %s" % (len(load), load)
            return 0
        
        while load:
            entry   = load[0:ENTRY_SIZE]
            load    = load[ENTRY_SIZE:]
            
            nid     = entry[0:self.nid_length]
            witness = entry[self.nid_length:2*self.nid_length]
            dist    = entry[2*self.nid_length:2*self.nid_length+2]
            seqn    = entry[-2:]

            
            result.append( [nid, witness, dist, seqn] )
            
        return result
        
    def get_weight(self, path):
        """ Calculates the weight of a given path from current node """
        cur = self.root
        w = 0
        for n in path:
            try:
                cur = cur.goto[n]
                w += cur.DIST  ############################# ADD WEIGHTS?????
            except:
                pout(0, "PANIC: weight: Cannot goto %s from %s" % (n, cur))
        return cur, w
    
    def follow_path(self, path):
        """ Encapsulation of get_weight """
        target, w = self.get_weight(path)
        return target
    
    def break_down(self):
        """ Breaks down the tree into node-witness pairs of the form:
        [node, witness, msg, dist]
        where msg is node's message and dist is the distance from current node"""
        
        for line in self.mst:
            node = line[0]
            path = line[1]
            cost = line[2]
            seqn = line[3]
            witness = self
            if path:
                witness = self.root.goto[path[0]].nid
            
            yield [node, witness, cost, seqn]
    
    def build_mindists(self):
        self.mst.clear()
        
        for n, path in self.root.preorder():
            # if there is no path in MST for this node,
            # or the existing path is longer
            if not n.nid in self.mst.node \
                or n.total_dist < self.mst.get_dist(n.nid):
                self.mst.set_path(n=n.nid, d=n.total_dist, p=path, s=n.seqn)
            
    def process_presence_helper( self, cur, next_node, cost, seqn, source ):
        if cur.nid != next_node:
            if next_node in cur.goto:
                if seqn > cur.goto[next_node].seqn:
                    cur.goto[next_node].seqn = seqn
                    cur.goto[next_node].last_update = time.time()
                    self.seqnos[next_node] = seqn
                    #print '*******UPDATED******', cur.goto[next_node].seqn
                else:
                    #if 
                    pass
                    #print 'Seqno is old', seqn, cur.goto[next_node].seqn
            else:
                cur.add_node(next_node, cost, seqn)
                #print 'added', bin2mac(next_node), cost, 'seqn', seqn
            cur = cur.goto[next_node]
        return cur
            
    def process_presence(self, entry, source):
        # make sure 'entry' is not empty
        if not len(entry): return 0
        
        # if I am the witness, skip it
        if entry[self.WITNESS] == self.nid:
            pout(1, " already know about ", entry)
            return 0
        
        # don't process presence about myself
        if entry[self.NID] == self.nid:
            pout(1, " this is about me ", entry)
            return 0
        
        # unpack distance to numeric form
        entry[self.DIST] = unpack("H", entry[self.DIST])[0]
        # unpack sequence number to numeric form
        entry[self.SEQN] = unpack("H", entry[self.SEQN])[0]
        seqn = 0

        # we only check the seqno if this entry is not about the source itself
        # otherwise, we always trust the seqno to be fresh
        if source != entry[self.NID]:
            if entry[self.NID] in self.seqnos:
                recved_no = entry[self.SEQN]
                stored_no = self.seqnos[entry[self.NID]]
                if self.seqno_is_old(recved_no, stored_no):
                    return 0
            #else:
                #print 'WILL probably ADD', bin2mac(entry[self.NID]), \
                    #entry[self.SEQN]

        cur = self.root
        next_node = source
        cost = self.T/self.interval
        if next_node == entry[self.NID]:
            seqn = entry[self.SEQN]
        cur = self.process_presence_helper( cur, next_node, cost, seqn, source )
        
        next_node = entry[self.WITNESS]
        cost = entry[self.DIST]
        if next_node == entry[self.NID]:
            seqn = entry[self.SEQN]
        cur = self.process_presence_helper( cur, next_node, cost, seqn, source )
        
        next_node = entry[self.NID]
        cost = entry[self.DIST]
        if next_node == entry[self.NID]:
            seqn = entry[self.SEQN]
        cur = self.process_presence_helper( cur, next_node, cost, seqn, source )
        
        #cur.dist = entry[self.DIST]
        
    def go_silent(self):
        self.silent = True
        
    def go_loud(self):
        self.silent = False
        
    def next_seqno(self):
        if self.sequenceno == SEQNO_MAX:
            self.sequenceno = 0
        else:
            self.sequenceno += 1
        return self.sequenceno

    def seqno_is_old(self, recved_no, stored_no):
        return recved_no <= stored_no and recved_no + SEQNO_MAX > stored_no + SEQNO_THRESHOLD

    def node_is_outofdate(self, node):
        # FIXME: I should be using arrival estimates here
        return time.time() - node.last_update > 10 * self.interval

    def do_maintainance(self):
        ### MAINTAIN tree
        now = time.time()
        
        for n in self.root.goto.values():
            # delete all arrivals older than T secs ago by
            # recursively checking if the first arrival is too old
            while len(n.arr_times) > self.T/self.interval \
                or (n.arr_times and n.arr_times[0] <= now - self.T):
                del n.arr_times[0]
                
        #log = {}
        #for w in self.root.goto.values():
            #log[w] = []
            #for n,p in w.preorder():
                #log[w].append([n.nid, p, n.dist, n.total_dist, len(n.arr_times)])
        
        for n in self.root.goto.values():
            if not n.arr_times:
                # delete the subtree rooted at that node with zero arrivals
                self.remove_subtree( n )
            else:
                for node, path in n.preorder():
                    if self.node_is_outofdate( node ):
                        #print bin2mac(n.nid), ": I THINK I SHOULD delete",\
                        #    bin2mac(node.nid), [bin2mac(x) for x in path]
                        self.remove_subtree( node )
                
        for neighbor in self.root.goto.values():
            # estimate next arrival based on some stochastic model
            self.estimate_cutoff_exp( neighbor )
                
        before = self.mst.node
        self.build_mindists()
        after  = self.mst.node
        
        ######## Take care of events
        _union     = set(before).union(self.mst.node)
        new_nodes  = _union - set(before)
        dept_nodes = _union - set(self.mst.node)
        if new_nodes and self.onNodeArrival:
            self.onNodeArrival([k for k in new_nodes])
        if dept_nodes and self.onNodeDeparture:
            self.onNodeDeparture([k for k in dept_nodes])
        
        
        ######## Take care of outstanding requests
        #self.teleporter.process_requests()
        
        
        ######## Take care of outstanding receptions
        #self.teleporter.process_receptions()
        
        
        ####### SEND FRAMES TO NEIGHBORS
        payload = ""
        if self.above_artificial_noise() and not self.silent:
            # Initialize payload by adding the present node
            payload = self.nid + self.nid + pack("H", 0) + pack('H', self.next_seqno())
            for e in self.break_down():
                try:
                    payload += e[0] + e[1] + pack("H", e[2]) + pack('H', e[3])
                    #print "****Will send about", bin2mac(e[0]), \
                     #   "witness", bin2mac(e[1]), e[3]
                except:
                    print "e123:",e[0], e[1], e[2], e[3]
                    raise
                    
            #sent = self.UDPsock.sendto( payload, self.other_addr )
            #pout(0, "sent %d bytes to interface" % sent)
        else:
            output += "did NOT send to %d\n" % n
            
        return payload
    
    def process_frame(self, addr, load):
        #self.pout(0, "data", load, len(load))
        entries = self.parse_entries( load )
        if not entries:
            return True
        
        for entry in entries:
            self.process_presence( entry, addr )
        
        # Add an arrival from this source
        try:
            self.root.goto[addr].add_arrival()
        except:
            print "WHY THE HELL is", bin2mac(addr), "not in", self.root.goto
            for e in entries:
                #for p in e[1]:
                print "entry:", bin2mac(e[0]), "path:", bin2mac(e[1]), "dist:", e[2]
            print "hmmm", bin2mac(self.nid), bin2mac(addr), self.nid == addr
            print

